package JSPservletPkg;

import java.io.*;
import javax.naming.*;
import javax.naming.directory.*;
import javax.naming.ldap.*;
import java.util.*;
import java.security.cert.*;

/**
 * CRL LDAP checker invoked by JSPloader for a given X509 certificate.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0
 * @see JSPservletPkg.JSPloader
 */
class CRLchecker {
  /** X509 Certificate Revocation List */
  private X509CRL crl = null;
  /** X509 CA Certificates */
  private X509Certificate[] caCerts = null;
  /** JSPhandler this CRLchecker has been created on behalf of */
  private JSPhandler handler;
  /**
   * Time of next check when CRL is null or CAURL is set but caCerts is null.
   * Its initial value (-2) is set to handle the construction case where
   * crl == caCerts == null.
   */
  private long nextCheck = -2;
  /**
   * Defines how often the LDAP server is queried in case of failure.
   * Used only when different signatures in the same archive.
   */
  private final static long scantime = 10000;  // 10 seconds
  /**
   * Constructor. Creates crl.
   * @param   handler JSPhandler
   */
  public CRLchecker(JSPhandler handler) {
    this.handler = handler;
  }
  /**
   * Read the CRL from the LDAP directory.
   * @return JSPloaderException raised when fails to get a CRL or if it is
   * not yet time to try again.
   */
  private JSPloaderException refresh() {
    if ((crl == null) || ((handler.CAURL != null) && (caCerts == null))) {
      long current = System.currentTimeMillis();
      if ((nextCheck != -2) && (current < nextCheck))
        return new JSPloaderException(handler, toString() +
          ".refresh Current crl:" + crl + " or caCerts:" + caCerts +
          " null, next attempt at " + new Date(nextCheck),
          JSPloaderException.JLE_LDAPACCESS);
      else
        nextCheck = current + scantime;
    }
    byte[] val = null;
    Hashtable env = new Hashtable();
    env.put(Context.INITIAL_CONTEXT_FACTORY, "com.sun.jndi.ldap.LdapCtxFactory");
    env.put(Context.PROVIDER_URL, handler.CRLURL);
    env.put(Context.REFERRAL, "follow");
    env.put(Context.SECURITY_AUTHENTICATION, "simple");
    if (handler.CRLLDAPuser != null) {
      env.put(Context.SECURITY_PRINCIPAL, handler.CRLLDAPuser);
      if (handler.CRLLDAPpasswd != null)
        env.put(Context.SECURITY_CREDENTIALS, handler.CRLLDAPpasswd);
    }
    JSPloaderException toThrow = null;
    CertificateFactory cf = null;
    try {
      cf = CertificateFactory.getInstance("X.509");
    }
    catch(CertificateException ce) {
      return new JSPloaderException(handler, toString() + ".refresh " + ce,
        JSPloaderException.JLE_LDAPACCESS);
    }
    try {
      DirContext ctx = new InitialDirContext(env);
      Attributes avals = ctx.getAttributes("");
      Attribute aval = avals.get("certificateRevocationList");
      val = (byte[])aval.get();
    }
    catch(Exception e) {
      toThrow = new JSPloaderException(handler, toString() +
        ".refresh Unchecked LDAP CRLURL:" + handler.CRLURL + " " + e,
        JSPloaderException.JLE_LDAPACCESS);
    }
    if ((toThrow != null) || (val == null) || (val.length == 0))
      toThrow = new JSPloaderException(handler, toString() +
        ".refresh Unchecked LDAP CRLURL:" + handler.CRLURL,
        JSPloaderException.JLE_LDAPACCESS);
    else {
      try {
        InputStream inStream = new ByteArrayInputStream(val);
        crl = (X509CRL)cf.generateCRL(inStream);
        inStream.close();
      }
      catch(Exception e) {
        toThrow = new JSPloaderException(handler, toString() +
          ".refresh Invalid LDAP CRLURL:" + handler.CRLURL + " " + e,
          JSPloaderException.JLE_LDAPCONFIG);
      }
    }
    if (handler.CAURL == null)
      return toThrow;
    env.put(Context.PROVIDER_URL, handler.CAURL);
    if (handler.CALDAPuser != null) {
      env.put(Context.SECURITY_PRINCIPAL, handler.CALDAPuser);
      if (handler.CALDAPpasswd != null)
        env.put(Context.SECURITY_CREDENTIALS, handler.CALDAPpasswd);
    }
    try {
      DirContext ctx = new InitialDirContext(env);
      Attributes avals = ctx.getAttributes("");
      Attribute aval = avals.get("cACertificate");
      val = (byte[])aval.get();
      if ((val != null) && (val.length != 0)) {
        InputStream inStream = new ByteArrayInputStream(val);
        Collection certificates = cf.generateCertificates(inStream);
        caCerts = (X509Certificate[])certificates.toArray();
        inStream.close();
      }
    }
    catch(Exception e) {
      handler.log.errorPrint(toString() +
        ".refresh Unable to retrieve caCerts: CAURL:" + handler.CAURL + " " + e);
    }
    handler.log.logprint(toString() + ".refresh url:" + handler.CRLURL + " user:"
      + handler.CRLLDAPuser + " password:" + handler.CRLLDAPpasswd);
    return toThrow;
  }
  /**
   * Checks a certificate is not on CRL and get next CRL update time.
   * @param   certs Certificate array to check.
   * @return long time since the epoch where next CRL update is scheduled.
   * @exception JSPloaderException raised when refresh failed to
   * get a CRL/caCert or when the certificate is revoked or the certificate
   * chain doesn't contain the caCert.
   */
  public long check(Certificate[] certs) throws JSPloaderException {
    JSPloaderException jle = refresh();
    if ((crl == null) && (caCerts == null))
      throw jle;
    boolean bValidated = false;
    long rc = 7 * 24 * 3600 * 1000;
    if (crl != null) {
      Date nextUpdate = crl.getNextUpdate();
      if (nextUpdate != null)
        rc = nextUpdate.getTime();
    }
    for (int j = 0; j < certs.length; ++j) {
      if (certs[j] instanceof X509Certificate) {
        X509Certificate xc = (X509Certificate)certs[j];
        int i = 0;
        if (caCerts != null) {
          for (; i < caCerts.length; ++i) {
            if (caCerts[i].equals(xc)) {
              bValidated = true;
              handler.log.logprint(toString() + ".check Issuer:" +
                xc.getIssuerDN() + " Serial:" + xc.getSerialNumber() +
                " found CA");
              break;
            }
          }
        }
        if ((caCerts == null) || (i == caCerts.length))
          handler.log.logprint(toString() + ".check Issuer:" + xc.getIssuerDN() +
            " Serial:" + xc.getSerialNumber());
        if (crl != null) {
          X509CRLEntry xce = crl.getRevokedCertificate(xc.getSerialNumber());
          if (xce != null)
            throw new JSPloaderException(handler,
              toString() + ".check revoked since:" + xce.getRevocationDate() +
              ", serial:" + xce.getSerialNumber(),
              JSPloaderException.JLE_REVOKED);
        }
      }
    }
    if ((caCerts != null) && !bValidated)
      throw new JSPloaderException(handler, toString() +
        ".check invalid signature", JSPloaderException.JLE_CERT);
    if (jle != null)
      throw jle;
    return rc;
  }
}
