package JSPservletPkg;
import java.util.*;
import java.net.*;
import java.io.*;

import java.security.*;

/**
 * Multicast support through ICP-like messages.
 * <p>Copyright (c) 2000-2001 Alexis Grandemange</p>
 * <pre>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; version
 * 2.1 of the License.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 * </pre>
 * <p>I included a copy of the GNU Lesser General Public License lesser.txt
 * in this directory.</p>
 * <p>alexis.grandemange@pagebox.net</p>
 * @author  Alexis Grandemange
 * @version 1, 0
 * @see JSPhandler.ClassEntry
 */
class ICPhandler extends Thread {
  /**
   * Maximum size of the ICP messages ICPhandler processes.
   * Size needed for properties.
   */
  static final int maxMsgSize = 16384;
  /**
   * A place holder to detect zero-filled or malformed messages.  A
   * cache must never intentionally send an ICP_OP_INVALID message.
   * ICP_OP_ERR should be used instead.
   */
  static final byte ICP_OP_INVALID = 0;
  /**
   * A query message. NOTE this opcode has a different payload format
   * than most of the others.  First is the requester's IPv4 address,
   * followed by a URL.  The Requester Host Address is not that of the
   * cache generating the ICP message, but rather the address of the
   * caches's client that originated the request.  The Requester Host
   * Address is often zero filled.  An ICP message with an all-zero
   * Requester Host Address address should be taken as one where the
   * requester address is not specified; it does not indicate a valid
   * IPv4 address.
   * In response to an ICP_OP_QUERY, the recipient must return one of:
   * ICP_OP_HIT, ICP_OP_MISS, ICP_OP_ERR, ICP_OP_MISS_NOFETCH,
   * ICP_OP_DENIED, or ICP_OP_HIT_OBJ.
   */
  static final byte ICP_OP_QUERY = 1;
  /**
   * An ICP_OP_HIT response indicates that the requested URL exists in
   * this cache and that the requester is allowed to retrieve it.
   */
  static final byte ICP_OP_HIT = 2;
  /**
   * An ICP_OP_MISS response indicates that the requested URL does not
   * exist in this cache.  The querying cache may still choose to fetch
   * the URL from the replying cache.
   */
  static final byte ICP_OP_MISS = 3;
  /**
   * An ICP_OP_ERR response indicates some kind of error in parsing or
   * handling the query message (e.g. invalid URL).
   */
  static final byte ICP_OP_ERR  = 4;
  // 5-9 UNUSED in standard ICP
  /**
   * PageBox initialization message issued by client.
   * Prompts for stats and archive names.
   */
  static final byte ICP_OP_INI  = 5;
  /**
   * PageBox archive RESP. Returned by server when it receives a ICP_OP_INI.
   * In this case, the payload is a serialized property file.
   */
  static final byte ICP_OP_ARCHIVE_STATUS  = 6;
  /** PageBox stats query issued by client. */
  static final byte ICP_OP_STAT_QUERY  = 7;
  /** PageBox log message. */
  static final byte ICP_OP_LOG  = 8;
  /** PageBox termination message. */
  static final byte ICP_OP_FIN  = 9;
  /**
   * Similar to ICP_OP_QUERY, but for use in simulating a query to an
   * origin server.  When ICP is used to select the closest neighbor,
   * the origin server can be included in the algorithm by bouncing an
   * ICP_OP_SECHO message off it's echo port.  The payload is simply
   * the null-terminated URL.
   */
  static final byte ICP_OP_SECHO = 10;
  /**
   * Similar to ICP_OP_QUERY, but for use in simulating a query to a
   * cache which does not use ICP.  When ICP is used to choose the
   * closest neighbor, a non-ICP cache can be included in the algorithm
   * by bouncing an ICP_OP_DECHO message off it's echo port.  The
   * payload is simply the null-terminated URL.
   */
  static final byte ICP_OP_DECHO = 11;
  // 12-20 UNUSED in standard ICP
  /**
   * PageBox stats response issued by server.
   * In this case, the payload is a serialized property file.
   */
  static final byte ICP_OP_STAT_STATUS  = 12;
  /**
   * PageBox update request issued by client.
   * In this case, the payload is a serialized property file.
   */
  static final byte ICP_OP_ARCHIVE_UPDATE = 13;
  /**
   * PageBox update request issued by client.
   * In this case, the payload is a serialized property file.
   */
  static final byte ICP_OP_STAT_UPDATE = 14;
  /**
   * PageBox publish message sent to Web Caches.
   */
  static final byte ICP_OP_ARCHIVE_PUBLISH = 15;
  /**
   * An ICP_OP_MISS_NOFETCH response indicates that this cache is up,
   * but is in a state where it does not want to handle cache misses.
   * An example of such a state is during a startup phase where a cache
   * might be rebuilding its object store.  A cache in such a mode may
   * wish to return ICP_OP_HIT for cache hits, but not ICP_OP_MISS for
   * misses.  ICP_OP_MISS_NOFETCH essentially means "I am up and
   * running, but please don't fetch this URL from me now."
   */
  static final byte ICP_OP_MISS_NOFETCH = 21;
  /**
   * An ICP_OP_DENIED response indicates that the querying site is not
   * allowed to retrieve the named object from this cache.  Caches and
   * proxies may implement complex access controls.  This reply must be
   * be interpreted to mean "you are not allowed to request this
   * particular URL from me at this particular time."
   */
  static final byte ICP_OP_DENIED = 22;
  /**
   * Just like an ICP_OP_HIT response, but the actual object data has
   * been included in this reply message.   Many requested objects are
   * small enough that it is possible to include them in the query
   * response and avoid the need to make a subsequent HTTP request for
   * the object.
   */
  static final byte ICP_OP_HIT_OBJ = 23;
  /** req number. */
  static int nextreq = 0;
  /** ICP header. */
  class IcpHeader {
  	/** opcode - unsigned in C. */
    byte opcode;
    /** version number - unsigned in C. */
    byte version = 2;
    /** total length (bytes) - unsigned in C. */
    short length;
    /** req number (req'd for UDP) - u_num32 in C. */
    int reqnum;
    /** flags - options in RFC 2186 - u_num32 in C. */
    byte[] flags = {0, 0, 0, 0};
    /** pad - option data in RFC 2186 - u_num32 in C. */
    byte[] pad = {0, 0, 0, 0};
    /** sender host id - u_num32 in C. */
    byte[] shostid = {0, 0, 0, 0};
    // PageBox specific.
    /** Timestamp. */
    long lastModified = 0;
    /** Identifier of requester. */
    byte[] ID;
    /** payload. */
    byte[] payload = null;
    /**
     * Constructor. Used on server side.
     * @param hostid sender host ID.
     * @param code OP code.
     * @param pl payload.
     * @param ID ID of requester. Can be handler.ID or handler.ID/jarName.
     * @param seqnum request number to use.
     * @param lastModified last modification time.
     */
    IcpHeader(byte[] hostid, byte code, byte[] pl, byte[] ID, int seqnum,
      long lastModified) {
      opcode = code;
      if (pl != null)
        length = (short)(pl.length + ID.length + 1 /* null separator */ + 20);
      else
        length = (short)(ID.length + 1 /* null separator */ + 20);
      payload = pl;
      shostid = hostid;
      this.ID = ID;
      this.lastModified = lastModified;
      reqnum = seqnum;
    }
    /**
     * Constructor. Used on client side.
     * @param hostid sender host ID.
     * @param code OP code.
     * @param pl payload.
     * @param ID ID of requester. Can be handler.ID or handler.ID/jarName.
     */
    IcpHeader(byte[] hostid, byte code, byte[] pl, byte[] ID) {
      opcode = code;
      if (pl != null)
        length = (short)(pl.length + ID.length + 1 /* null separator */ + 20);
      else
        length = (short)(ID.length + 1 /* null separator */ + 20);
      payload = pl;
      shostid = hostid;
      this.ID = ID;
      synchronized(IcpHeader.class) {
        reqnum = nextreq ++;
      }
    }
    /**
     * Constructor.
     * Populated from a buffer
     * @param msg buffer.
     */
    IcpHeader(byte[] msg) {
      opcode = msg[0];
      version = msg[1];
      length = (short)((ICPhandler.toInt(msg[2]) << 8) + ICPhandler.toInt(msg[3]));
      reqnum = (ICPhandler.toInt(msg[4]) << 24) + (ICPhandler.toInt(msg[5]) <<
        16) + (ICPhandler.toInt(msg[6]) << 8) + ICPhandler.toInt(msg[7]);
      System.arraycopy(msg, 8, flags, 0, 4);
      System.arraycopy(msg, 12, pad, 0, 4);
      int i = 0;
      for (; i < 8; ++i)
        lastModified = (lastModified << 8) + ICPhandler.toInt(msg[0 + i]);
      System.arraycopy(msg, 16, shostid, 0, 4);
      for (i = 0; i < length - 20; ++i) {
        if (msg[20 + i] == 0)
          break;
      }
      ID = new byte[i];
      System.arraycopy(msg, 20, ID, 0, i);
      if (i == length - 21)
        return;
      payload = new byte[length - 21 -i];
      System.arraycopy(msg, 21 + i, payload, 0, length - 21 -i);
    }
    /** Returns serialized length. */
    int length() {
      return length;
    }
    /** Returns serialized flow. */
    byte[] getBytes() {
      byte[] newMsg = new byte[length];
      newMsg[0] = opcode;
      newMsg[1] = version;
      newMsg[2] = (byte)(length >>> 8);
      newMsg[3] = (byte)(length & 0x00ff);
      newMsg[4] = (byte)(reqnum >>> 24);
      newMsg[5] = (byte)((reqnum >>> 16) & 0x000000ff);
      newMsg[6] = (byte)((reqnum >>> 8) & 0x000000ff);
      newMsg[7] = (byte)(reqnum & 0x000000ff);
      if (lastModified == 0) {
        System.arraycopy(flags, 0, newMsg, 8, 4);
        System.arraycopy(pad, 0, newMsg, 12, 4);
      } else {
        newMsg[8] = (byte)(lastModified >>> 56);
        newMsg[9] = (byte)((lastModified >>> 48) & 0x00000000000000ff);
        newMsg[10] = (byte)((lastModified >>> 40) & 0x00000000000000ff);
        newMsg[11] = (byte)((lastModified >>> 32) & 0x00000000000000ff);
        newMsg[12] = (byte)((lastModified >>> 24) & 0x00000000000000ff);
        newMsg[13] = (byte)((lastModified >>> 16) & 0x00000000000000ff);
        newMsg[14] = (byte)((lastModified >>> 8) & 0x00000000000000ff);
        newMsg[15] = (byte)(lastModified & 0x00000000000000ff);
      }
      System.arraycopy(shostid, 0, newMsg, 16, 4);
      System.arraycopy(ID, 0, newMsg, 20, ID.length);
      newMsg[20 + ID.length] = 0;
      if (payload != null)
        System.arraycopy(payload, 0, newMsg, 20 + ID.length + 1, payload.length);
      return newMsg;
    }
  }
  /** Multicast socket. */
  private MulticastSocket mcs = null;
  /** Multicast group or point to point interface. */
  private InetAddress group = null;
  /** UDP point to point socket */
  private DatagramSocket ds = null;
  /** AS host address. */
  private byte[] myAddress = null;
  /** Invoking JSPhandler. */
  private JSPhandler handler;
  /** Loop while true. */
  private boolean toContinue = true;
  /** Check if loop must be continued. */
  private synchronized boolean toBeContinued() {
    return toContinue;
  }
  /**
   * Converts a byte in positive integer.
   * @param b byte to convert.
   */
  final static int toInt(byte b) {
    int i = b;
    if (i < 0)
      return 256 + i;
    return i;
  }
  static void dump(byte[] buf, int length) {
    String msg = "";
    int i = 0, cnt = 0;
    for (cnt = 0; i < ((length / 4) * 4); i += 4) {
      if (cnt == 5) {
        msg += Integer.toHexString((toInt(buf[i]) << 24) + (toInt(buf[i + 1]) <<
          16) + (toInt(buf[i + 2]) << 8) + toInt(buf[i + 3])) + "\n";
        cnt = 0;
      } else {
        msg += Integer.toHexString((toInt(buf[i]) << 24) + (toInt(buf[i + 1]) <<
          16) + (toInt(buf[i + 2]) << 8) + toInt(buf[i + 3])) + " ";
        ++cnt;
      }
    }
    int lastWord = 0;
    for (; i < length; ++i)
      lastWord = (lastWord << 8) + toInt(buf[i]);
    msg += Integer.toHexString(lastWord);
    System.out.println(msg + "length=" + length);
  }
  /**
   * Stop the loop.<OL>
   * <LI>Set toContinue at false to ask thread termination.
   * <LI>sends a message. On client side, triggers the termination.
   * On server size, triggers log close.</LI></OL>
   */
  private void endLoop() {
    synchronized(this) {
      toContinue = false;
    }
    try {
      IcpHeader ih = new IcpHeader(myAddress, ICP_OP_FIN, null,
        handler.ID.getBytes("ISO-8859-1"));
      if (handler.bMulticast) {
        DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(), group,
          handler.port);
        mcs.send(hi);
      } else {
        if (handler.bDisklessClt) {
          for (int i = 0; i < remAddrs.length; ++i) {
            DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(),
              remAddrs[i], handler.remPorts[i]);
            ds.send(hi);
          }
        }
        DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(), group,
          handler.port);
        ds.send(hi);
      }
    }
    catch(UnsupportedEncodingException uee) {
      System.out.println(toString() + ".endLoop " + uee);
    }
    catch(IOException ioe) {
      System.out.println(toString() + ".endLoop " + ioe);
    }
  }
  InetAddress remAddrs[] = null;
  /**
   * Constructor.
   * Log is created after ICPhandler. Do not use handler.log.
   * @param jh calling JSPhandler.
   */
  ICPhandler(JSPhandler jh) throws org.osgi.framework.BundleException {
    handler = jh;
    String step = "MulticastSocket";
    try {
      if (handler.mcg != null)
        group = InetAddress.getByName(handler.mcg);
      if (handler.bMulticast) {
        mcs = new MulticastSocket(handler.port);
        mcs.setSoTimeout(1000 /* 1 sec */);
        step = "setSoTimeout";
        mcs.joinGroup(group);
      } else {
        if (group == null)
          ds = new DatagramSocket(handler.port);
        else
          ds = new DatagramSocket(handler.port, group);
        ds.setSoTimeout(1000 /* 1 sec */);
        if (handler.remAddrs != null) {
          remAddrs = new InetAddress[handler.remAddrs.length];
          for (int i = 0; i < handler.remAddrs.length; ++i)
            remAddrs[i] = InetAddress.getByName(handler.remAddrs[i]);
        }
      }
      myAddress = InetAddress.getLocalHost().getAddress();
    }
    catch(Exception e) {
      throw new org.osgi.framework.BundleException(toString() + ".ICPhandler(" +
        handler.mcg + ", " + handler.port + ", " + myAddress + ") " + step + " "
        + e, e);
    }
    if (handler.toTrace) {
      String szDiskless = handler.bDisklessClt ? ") client" : ") server";
      String szMulticast = handler.bMulticast ? " multicast" : " unicast";
      System.out.println(toString() + ".ICPhandler(" + handler.mcg + ", " +
        handler.port + ", " + toInt(myAddress[0]) + "." + toInt(myAddress[1]) +
        "." + toInt(myAddress[2]) + "." + toInt(myAddress[3]) + szDiskless +
        szMulticast);
    }
    if (handler.bDisklessClt) {
      if (!handler.bMulticast && handler.toTrace) {
        String lines = "";
        for (int i = 0; i < remAddrs.length; ++i) {
          byte[] b = remAddrs[i].getAddress();
          lines += "\n\taddr=" + b[0] + "." + b[1] + "." + b[2] + "." + b[3] +
            " port=" + handler.remPorts[i];
        }
        System.out.println(toString() + ".ICPhandler multicast" + lines);
      }
      disklessClientInit();
    } else {
      logs = new HashMap();
      dch = new DisklessCltHandler(handler);
    }
    start();
  }
  /**
   * Map of found archive stats on diskless client side.
   * Used by ClassEntry.Stat.
   * Key: archive name.
   * Value: stat properties.
   */
  HashMap stats = new HashMap();
  /**
   * On client side - not use handler.log.
   * Diskless client handshaking - synchronous.<UL>
   * <LI>==> ICP_OP_INI</LI>
   * <LI><== ICP_OP_ARCHIVE_STATUS</LI>
   * <LI>==> ICP_OP_STAT_QUERY</LI>
   * <LI><== ICP_OP_STAT_STATUS</LI></UL>
   * <pre>Suppose two server answer whose one was stopped during last client session.
   * Client must give a chance to all up servers - though it doesn't know how
   * many are up and select the most up to date answer.
   */
  public void disklessClientInit() throws org.osgi.framework.BundleException {
    try {
      IcpHeader sendIh = new IcpHeader(myAddress, ICP_OP_INI, null,
        handler.ID.getBytes("ISO-8859-1"));
      int reqnum = sendIh.reqnum;
      if (handler.bMulticast) {
        DatagramPacket hi = new DatagramPacket(sendIh.getBytes(),
          sendIh.length(), group, handler.port);
        mcs.send(hi);
      } else {
        for (int i = 0; i < remAddrs.length; ++i) {
          DatagramPacket hi = new DatagramPacket(sendIh.getBytes(),
            sendIh.length(), remAddrs[i], handler.remPorts[i]);
          ds.send(hi);
        }
      }
      if (handler.toTrace)
        System.out.println(toString() + ".disklessClientInit(" + handler.ID +
          ") sent ICP_OP_INI req=" + reqnum);
      long cur = System.currentTimeMillis();
      long retryTime = cur + 5000;
      long decisionTime = cur + 1000;
      long timeToBeat = 0;
      IcpHeader ihToBeat = null, receiveIh = null;
      while(true) {
        cur = System.currentTimeMillis();
        if ((cur > decisionTime) && (ihToBeat != null)) {
          if (ihToBeat.payload == null)
            break;
          String nameVal = handler.disklessRemotePropInit(
            new ByteArrayInputStream(ihToBeat.payload));
          if (handler.toTrace)
            System.out.println(toString() + ".disklessClientInit(" + handler.ID
              + ") remoteLocProp" + nameVal);
          break;
        }
        if (cur > retryTime) {
          if (handler.bMulticast) {
            DatagramPacket hi = new DatagramPacket(sendIh.getBytes(),
              sendIh.length(), group, handler.port);
            mcs.send(hi);
          } else {
            for (int i = 0; i < remAddrs.length; ++i) {
              DatagramPacket hi = new DatagramPacket(sendIh.getBytes(),
                sendIh.length(), remAddrs[i], handler.remPorts[i]);
              ds.send(hi);
            }
          }
          retryTime = cur + 5000;
          decisionTime = cur + 1000;
        }
        DatagramPacket recv = new DatagramPacket(new byte[maxMsgSize],
          maxMsgSize);
        try {
          if (handler.bMulticast)
            mcs.receive(recv);
          else
            ds.receive(recv);
        }
        catch(InterruptedIOException iioe) {
          System.out.println(toString() + ".disklessClientInit(" + handler.ID +
            ") timeout expired for ICP_OP_INI");
          continue;
        }
        receiveIh = new IcpHeader(recv.getData());
        String ID = new String(receiveIh.ID);
        if (handler.toTrace)
          System.out.println(toString() + ".disklessClientInit(" + handler.ID +
            ") received " + receiveIh.opcode + "ID=" + ID + " req=" +
            receiveIh.reqnum + " from=" + toInt(receiveIh.shostid[0]) + "." +
            toInt(receiveIh.shostid[1]) + "." + toInt(receiveIh.shostid[2]) +
            "." + toInt(receiveIh.shostid[3]) + " lastModified=" +
            receiveIh.lastModified);
        if (receiveIh.opcode != ICP_OP_ARCHIVE_STATUS)
          continue;
        if (!ID.equals(handler.ID))
          continue;
        if (receiveIh.reqnum != reqnum)
          continue;
        if ((ihToBeat == null) ||
          (receiveIh.lastModified > ihToBeat.lastModified))
          ihToBeat = receiveIh;
      }
      if (handler.remoteLocProp.isEmpty())
        return;
      Enumeration e = handler.remoteLocProp.keys();
      while(e.hasMoreElements()) {
        String jarName = (String)e.nextElement();
        String idJar = handler.ID + "/" + jarName;
        sendIh = new IcpHeader(myAddress, ICP_OP_STAT_QUERY, null,
          idJar.getBytes("ISO-8859-1"));
        reqnum = sendIh.reqnum;
        if (handler.bMulticast) {
          DatagramPacket hi = new DatagramPacket(sendIh.getBytes(),
            sendIh.length(), group, handler.port);
          mcs.send(hi);
        } else {
          for (int i = 0; i < remAddrs.length; ++i) {
            DatagramPacket hi = new DatagramPacket(sendIh.getBytes(),
              sendIh.length(), remAddrs[i], handler.remPorts[i]);
            ds.send(hi);
          }
        }
        if (handler.toTrace)
          System.out.println(toString() + ".disklessClientInit(" + idJar +
            ") sent ICP_OP_STAT_QUERY req=" + reqnum);
        cur = System.currentTimeMillis();
        retryTime = cur + 5000;
        decisionTime = cur + 1000;
        timeToBeat = 0;
        ihToBeat = null;
        while(true) {
          cur = System.currentTimeMillis();
          if ((cur > decisionTime) && (ihToBeat != null)) {
            if (ihToBeat.payload == null)
              break;
            Properties p = new Properties();
            p.load(new ByteArrayInputStream(ihToBeat.payload));
            stats.put(jarName, p);
            if (handler.toTrace && !p.isEmpty()) {
              String nameVal = "";
              Enumeration en = p.keys();
              while(en.hasMoreElements()) {
                String key = (String)en.nextElement();
                nameVal += "\n\t" + key + "=" + p.getProperty(key);
              }
              System.out.println(toString() + ".disklessClientInit(" +
                handler.ID + ") stats" + nameVal);
            }
            break;
          }
          if (cur > retryTime) {
            if (handler.bMulticast) {
              DatagramPacket hi = new DatagramPacket(sendIh.getBytes(),
                sendIh.length(), group, handler.port);
              mcs.send(hi);
            } else {
              for (int i = 0; i < remAddrs.length; ++i) {
                DatagramPacket hi = new DatagramPacket(sendIh.getBytes(),
                  sendIh.length(), remAddrs[i], handler.remPorts[i]);
                ds.send(hi);
              }
            }
            retryTime = cur + 5000;
            decisionTime = cur + 1000;
          }
          DatagramPacket recv = new DatagramPacket(new byte[maxMsgSize],
            maxMsgSize);
          try {
            if (handler.bMulticast)
              mcs.receive(recv);
            else
              ds.receive(recv);
          }
          catch(InterruptedIOException iioe) {
            System.out.println(toString() + ".disklessClientInit(" + handler.ID
              + ") timeout expired for ICP_OP_STAT_QUERY");
            continue;
          }
          receiveIh = new IcpHeader(recv.getData());
          String ID = new String(receiveIh.ID);
          if (handler.toTrace)
            System.out.println(toString() + ".disklessClientInit(" + handler.ID
              + ") received 2 " + receiveIh.opcode + "ID=" + ID + " req=" +
              receiveIh.reqnum + " from=" + toInt(receiveIh.shostid[0]) + "." +
              toInt(receiveIh.shostid[1]) + "." + toInt(receiveIh.shostid[2]) +
              "." + toInt(receiveIh.shostid[3]) + " lastModified=" +
              receiveIh.lastModified);
          if (receiveIh.opcode != ICP_OP_STAT_STATUS)
            continue;
          if (!ID.equals(idJar))
            continue;
          if (receiveIh.reqnum != reqnum)
            continue;
          if ((ihToBeat == null) ||
            (receiveIh.lastModified > ihToBeat.lastModified))
            ihToBeat = receiveIh;
        }
      }
    }
    catch(UnsupportedEncodingException uee) {
      throw new org.osgi.framework.BundleException(toString() +
        ".disklessClientInit(" + handler.ID + ")", uee);
    }
    catch(IOException ioe) {
      throw new org.osgi.framework.BundleException(toString() +
        ".disklessClientInit(" + handler.ID + ")", ioe);
    }
  }
  /**
   * Loops to read multicast messages.
   */
  public void run() {
    if (handler.bDisklessClt) {
      while(toBeContinued())
        disklessClientCheck();
    } else {
      while(toBeContinued())
        disklessServerCheck();
    }
    try {
      if (handler.bMulticast) {
        mcs.leaveGroup(group);
        mcs.close();
      } else
        ds.close();
    }
    catch(Exception e) {
      if (handler.bDisklessClt)
        System.out.println(toString() + ".run " + e);
      else
        handler.log.errorPrint(toString() + ".run " + e);
    }
  }
  /**
   * Handles multicast messages targeting client.
   */
  private void disklessClientCheck() {
    DatagramPacket recv = new DatagramPacket(new byte[maxMsgSize], maxMsgSize);
    try {
      if (handler.bMulticast)
        mcs.receive(recv);
      else
        ds.receive(recv);
    }
    catch(InterruptedIOException iioe) {
      return;
    }
    catch(Exception e) {
      System.out.println(toString() + ".disklessClientCheck() " + e);
      return;
    }
  }
  /**
   * Class handling diskless IDs. On diskless server side.
   */
  class DisklessCltHandler {
    /**
     * Diskless ID set.
     */
    private Set handlers = null;
    /** Creating handler. */
    private JSPhandler handler;
    /** Serialization file. */
    private File f = null;
    /**
     * Constructor.
     * @param jh back pointer on creating JSPhandler.
     */
    DisklessCltHandler(JSPhandler jh) {
      handler = jh;
      try {
        f = handler.getFile(handler.cachePath + "/" + "diskless.ser");
        ObjectInputStream ois = null;
        if ((f != null) && f.exists())
          ois = new ObjectInputStream(new FileInputStream(f));
        if (ois == null)
          handlers = new HashSet();
        else
          handlers = (HashSet)ois.readObject();
      }
      catch(Exception e) {
        handler.log.errorPrint(toString() + ".DisklessCltHandler " + e);
      }
    }
    synchronized String[] getIDs() {
      if (handlers == null)
        return null;
      Object[] a = handlers.toArray();
      String[] rc = new String[a.length];
      String msg = toString() + ".getIDs found: ";
      for (int i = 0; i < a.length; ++i) {
        rc[i] = (String)a[i];
        msg += rc[i];
      }
      handler.log.logprint(msg);
      return rc;
    }
    synchronized void addID(String ID) {
      if (handlers == null)
        return;
      if (handlers.contains(ID))
        return;
      handler.log.logprint(toString() + ".addID(" + ID + ")");
      handlers.add(ID);
    }
    /** Termination method. Persists handlers. */
    void close() {
      if (f == null)
        return;
      try {
        ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(f));
        oos.writeObject(handlers);
      }
      catch(Exception e) {
        handler.log.errorPrint(toString() + ".close " + e);
      }
    }
    /** to avoid loitering */
    protected void finalize() {
      close();
    }
  }
  /** Object handling diskless IDs. Instantiated on diskless server side. */
  DisklessCltHandler dch = null;
  /**
   * Class wraping diskless logging. On diskless server side.
   */
  class Log {
    /** PrintWriter initialized by constructor. */
    private PrintWriter log = null;
    /** Back pointer on creating JSPhandler. */
    private JSPhandler handler;
    /** File on log. Used by DisklessLog. */
    File logFile = null;
    /**
     * Constructor.
     * @param jh back pointer on creating JSPhandler.
     */
    Log(String ID, JSPhandler jh) {
      handler = jh;
      try {
        logFile = handler.getFile(handler.cachePath + "/" + ID + ".txt");
        if (logFile != null)
          log = new PrintWriter(new BufferedWriter(new FileWriter(logFile)));
      }
      catch(Exception e) {
        System.out.println(toString() + ".Log " + e);
        log = null;
      }
    }
    /**
     * Logging method.
     * @param   msg message to format and print.
     */
    final void log(String msg) {
      if (log == null)
        return;
      log.println(msg);
      log.flush();
    }
    /** Empties the log file. */
    final synchronized void clear() {
      if (log == null)
        return;
      log.close();
      try {
        if (logFile != null) {
          if (logFile.exists())
            logFile.delete();
          log = new PrintWriter(new BufferedWriter(new FileWriter(logFile)));
        }
      }
      catch(Exception e) {
        log = null;
      }
    }
    /** Closes log file. */
    final void close() {
      if (log == null)
        return;
      log.close();
      log = null;
    }
    /** to avoid loitering */
    protected void finalize() {
      close();
    }
  }
  /**
   * Map of diskless logs. Used on diskless server side.
   * key: diskless ID
   * value: Log
   */
  private HashMap logs = null;
  /**
   * Log retrieve method. Called by disklessServerCheck and DisklessLog.
   * @param   ID logs key.
   */
  synchronized Log getLog(String ID) {
    Log log = (Log)logs.get(ID);
    if (log != null)
      return log;
    log = new Log(ID, handler);
    logs.put(ID, log);
    return log;
  }
  /**
   * Handle multicast messages targeting server.
   */
  private void disklessServerCheck() {
    DatagramPacket recv = new DatagramPacket(new byte[maxMsgSize], maxMsgSize);
    try {
      if (handler.bMulticast)
        mcs.receive(recv);
      else {
        ds.receive(recv);
      }
    }
    catch(InterruptedIOException iioe) {
      return;
    }
    catch(IOException ioe) {
      handler.log.errorPrint(toString() + ".disklessServerCheck() " + ioe);
      return;
    }
    InetAddress remIa = recv.getAddress();
    int remPort = recv.getPort();
    byte[] netbuf = recv.getData();
    // ICPhandler.dump(netbuf, recv.getLength());
    IcpHeader ih = new IcpHeader(netbuf);
    if (ih.opcode == ICP_OP_LOG) {
      String ID = new String(ih.ID);
      dch.addID(ID);
      Log clientLog = getLog(ID);
      /*
      handler.log.logprint(toString() +
        ".disklessServerCheck received ICP_OP_LOG ID=" + new String(ih.ID) +
        " req=" + ih.reqnum + " from=" + toInt(ih.shostid[0]) + "." +
        toInt(ih.shostid[1]) + "." + toInt(ih.shostid[2]) + "." +
        toInt(ih.shostid[3]) + " size=" + ih.length + " payload=" +
        ih.payload.length);
      */
      clientLog.log(new String(ih.payload));
    } else
    if (ih.opcode == ICP_OP_FIN) {
      String ID = new String(ih.ID);
      if (!ID.equals(handler.ID)) {
        Log clientLog = (Log)logs.get(ID);
        handler.log.logprint(toString() +
          ".disklessServerCheck received ICP_OP_FIN ID=" + new String(ih.ID) +
          " req=" + ih.reqnum + " from=" + toInt(ih.shostid[0]) + "." +
          toInt(ih.shostid[1]) + "." + toInt(ih.shostid[2]) + "." +
          toInt(ih.shostid[3]));
        if (clientLog != null)
          clientLog.close();
        logs.remove(ID);
      }
    } else
    if (ih.opcode == ICP_OP_INI) {
      String ID = new String(ih.ID);
      dch.addID(ID);
      handler.log.logprint(toString() +
        ".disklessServerCheck received ICP_OP_INI ID=" + new String(ih.ID) +
        " req=" + ih.reqnum + " from=" + toInt(ih.shostid[0]) + "." +
        toInt(ih.shostid[1]) + "." + toInt(ih.shostid[2]) + "." +
        toInt(ih.shostid[3]));
      try {
        File f = handler.getFile(handler.cachePath + "/" + ID + ".properties");
        if ((f != null) && f.exists()) {
          try {
            BufferedInputStream bis = new BufferedInputStream(new FileInputStream(f));
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            int i;
            while((i = bis.read()) != -1)
              baos.write(i);
            byte[] buf = baos.toByteArray();
            ih = new IcpHeader(myAddress, ICP_OP_ARCHIVE_STATUS, buf, ih.ID,
              ih.reqnum, f.lastModified());
          }
          catch(IOException ioe) {
            ih = new IcpHeader(myAddress, ICP_OP_ARCHIVE_STATUS, null, ih.ID,
              ih.reqnum, 0L);
          }
        } else
          ih = new IcpHeader(myAddress, ICP_OP_ARCHIVE_STATUS, null, ih.ID,
            ih.reqnum, 0L);
        // ICPhandler.dump(ih.getBytes(), ih.length());
        if (handler.bMulticast) {
          DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(),
            group, handler.port);
          mcs.send(hi);
        } else {
          DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(),
            remIa, remPort);
          ds.send(hi);
        }
        handler.log.logprint(toString() +
          ".disklessServerCheck sent ICP_OP_ARCHIVE_STATUS ID=" +
          new String(ih.ID));
      }
      catch(Exception e) {
        handler.log.errorPrint(toString() + ".disklessServerCheck " + e);
      }
    } else
    if (ih.opcode == ICP_OP_STAT_QUERY) {
      String jarId = new String(ih.ID);
      handler.log.logprint(toString() +
        ".disklessServerCheck received ICP_OP_STAT_QUERY ID=" + jarId + " req="
        + ih.reqnum + " from=" + toInt(ih.shostid[0]) + "." +
        toInt(ih.shostid[1]) + "." + toInt(ih.shostid[2]) + "." +
        toInt(ih.shostid[3]));
      try {
        File f = handler.getFile(handler.cachePath + "/" + jarId + ".stat");
        if ((f != null) && f.exists()) {
          try {
            BufferedInputStream bis = new BufferedInputStream(
              new FileInputStream(f));
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            int i;
            while((i = bis.read()) != -1)
              baos.write(i);
            byte[] buf = baos.toByteArray();
            ih = new IcpHeader(myAddress, ICP_OP_STAT_STATUS, buf, ih.ID,
              ih.reqnum, f.lastModified());
          }
          catch(IOException ioe) {
            ih = new IcpHeader(myAddress, ICP_OP_STAT_STATUS, null, ih.ID,
              ih.reqnum, 0L);
          }
        } else
          ih = new IcpHeader(myAddress, ICP_OP_STAT_STATUS, null, ih.ID,
            ih.reqnum, 0L);
        if (handler.bMulticast) {
          DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(),
            group, handler.port);
          mcs.send(hi);
        } else {
          DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(),
            remIa, remPort);
          ds.send(hi);
        }
        handler.log.logprint(toString() +
          ".disklessServerCheck sent ICP_OP_STAT_STATUS ID=" + new String(ih.ID));
      }
      catch(Exception e) {
        handler.log.errorPrint(toString() + ".disklessServerCheck " + e);
      }
    } else
    if ((ih.opcode == ICP_OP_ARCHIVE_UPDATE) && (ih.payload != null)) {
      String ID = new String(ih.ID);
      dch.addID(ID);
      handler.log.logprint(toString() +
        ".disklessServerCheck received ICP_OP_ARCHIVE_UPDATE ID=" + ID + " req="
        + ih.reqnum + " from=" + toInt(ih.shostid[0]) + "." +
        toInt(ih.shostid[1]) + "." + toInt(ih.shostid[2]) + "." +
        toInt(ih.shostid[3]));
      try {
        Properties clientRemoteLocProp = new Properties();
        clientRemoteLocProp.load(new DataInputStream(new ByteArrayInputStream(
          ih.payload)));
        File f = handler.getFile(handler.cachePath + "/" + ID + ".properties");
        BufferedOutputStream clientBos = new BufferedOutputStream(
          new FileOutputStream(f));
        clientRemoteLocProp.store(clientBos, null);
        clientBos.close();
      }
      catch(Exception e) {
        handler.log.errorPrint(toString() + ".disklessServerCheck " + e);
      }
    } else
    if ((ih.opcode == ICP_OP_STAT_UPDATE) && (ih.payload != null)) {
      String jarId = new String(ih.ID);
      /*
      handler.log.logprint(toString() +
        ".disklessServerCheck received ICP_OP_STAT_UPDATE ID=" + jarId + " req="
        + ih.reqnum + " from=" + toInt(ih.shostid[0]) + "." +
        toInt(ih.shostid[1]) + "." + toInt(ih.shostid[2]) + "." +
        toInt(ih.shostid[3]));
      */
      try {
        Properties clientStatProp = new Properties();
        clientStatProp.load(new DataInputStream(new ByteArrayInputStream(
          ih.payload)));
        File f = handler.getFile(handler.cachePath + "/" + jarId + ".stat");
        BufferedOutputStream clientBos = new BufferedOutputStream(
          new FileOutputStream(f));
        clientStatProp.store(clientBos, null);
        clientBos.close();
      }
      catch(Exception e) {
        handler.log.errorPrint(toString() + ".disklessServerCheck " + e);
      }
    } else
    if (!handler.bMulticast || ((ih.opcode != ICP_OP_ARCHIVE_STATUS) &&
      (ih.opcode != ICP_OP_STAT_STATUS)))
      handler.log.errorPrint(toString() + ".disklessServerCheck received " +
        ih.opcode + " ID=" + new String(ih.ID) + " req=" + ih.reqnum + " from="
        + toInt(ih.shostid[0]) + "." + toInt(ih.shostid[1]) + "." +
        toInt(ih.shostid[2]) + "." + toInt(ih.shostid[3]));
  }
  /**
   * Wraps a log.
   * Must be invoked by AccessController.doPrivileged.
   */
  class LogPrivilegedAction implements PrivilegedAction {
    /** Back pointer on ICPhandler. */
    ICPhandler icp;
    /** Message to Log. */
    String msg;
    /**
     * Constructor.
     * @param     ih invoking ICPhandler.
     * @param     msg message to log.
     */
    LogPrivilegedAction(ICPhandler ih, String msg) {
      icp = ih;
      this.msg = msg;
    }
    /** Invoked by AccessController.doPrivileged. */
    public Object run() {
      try {
        IcpHeader ih = new IcpHeader(icp.myAddress, ICP_OP_LOG, msg.getBytes(
          "ISO-8859-1"), icp.handler.ID.getBytes("ISO-8859-1"));
        // ICPhandler.dump(ih.getBytes(), ih.length());
        if (icp.handler.bMulticast) {
          DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(),
            group, icp.handler.port);
          icp.mcs.send(hi);
        } else {
          for (int i = 0; i < remAddrs.length; ++i) {
            DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(),
              remAddrs[i], icp.handler.remPorts[i]);
            icp.ds.send(hi);
          }
        }
      }
      catch(UnsupportedEncodingException uee) {
        System.out.println(toString() + ".log " + uee);
      }
      catch(IOException ioe) {
        System.out.println(toString() + ".log " + ioe);
      }
      return null;
    }
  }
  /**
   * Log method - used by diskless client.
   * In case of exceptions, sends messages on console.
   * @param msg message to log.
   */
  void log(String msg) {
    AccessController.doPrivileged(new LogPrivilegedAction(this, msg));
  }
  /**
   * Archive publication sent to Web caches.
   */
  synchronized void publish() {
    if (handler.classEntries.isEmpty())
      return;
    try {
      Set archiveSet = handler.classEntries.keySet();
      Iterator it = archiveSet.iterator();
      byte[] myurl = handler.myurl.getBytes("ISO-8859-1");
      byte[] context = handler.contextPath.getBytes("ISO-8859-1");
      int bufsz = myurl.length + context.length + 2;
      byte[][] ba = new byte[archiveSet.size()][];
      int idx = 0;
      for(; it.hasNext(); ++ idx) {
        String published = (String)it.next();
        ba[idx] = published.getBytes("ISO-8859-1");
        bufsz += (ba[idx].length + 1);
      }
      byte[] buf = new byte[bufsz];
      System.arraycopy(myurl, 0, buf, 0, myurl.length);
      buf[myurl.length] = '\0';
      int pos = myurl.length + 1;
      System.arraycopy(context, 0, buf, pos, context.length);
      pos += context.length;
      buf[pos ++] = '\0';
      for (int i = 0; i < idx; ++ i) {
        System.arraycopy(ba[i], 0, buf, pos, ba[i].length);
        pos += ba[i].length;
        buf[pos ++] = '\0';
      }
      IcpHeader ih = new IcpHeader(myAddress, ICP_OP_ARCHIVE_PUBLISH, buf,
        handler.ID.getBytes("ISO-8859-1"));
      ICPhandler.dump(ih.getBytes(), ih.length());
      if (handler.bMulticast) {
        DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(),
          group, handler.port);
        mcs.send(hi);
      } else {
        for (int i = 0; i < remAddrs.length; ++i) {
          DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(),
            remAddrs[i], handler.remPorts[i]);
          ds.send(hi);
        }
      }
    }
    catch(UnsupportedEncodingException uee) {
      System.out.println(toString() + ".publish " + uee);
    }
    catch(IOException ioe) {
      System.out.println(toString() + ".publish " + ioe);
    }
  }
  /**
   * Update method - used by diskless client.
   * In case of exceptions, sends messages on console.
   */
  synchronized void update() {
    try {
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      handler.remoteLocProp.store(baos, null);
      byte[] buf = baos.toByteArray();
      IcpHeader ih = new IcpHeader(myAddress, ICP_OP_ARCHIVE_UPDATE, buf,
        handler.ID.getBytes("ISO-8859-1"));
      // ICPhandler.dump(ih.getBytes(), ih.length());
      if (handler.bMulticast) {
        DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(), group,
          handler.port);
        mcs.send(hi);
      } else {
        for (int i = 0; i < remAddrs.length; ++i) {
          DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(),
            remAddrs[i], handler.remPorts[i]);
          ds.send(hi);
        }
      }
    }
    catch(UnsupportedEncodingException uee) {
      System.out.println(toString() + ".update " + uee);
    }
    catch(IOException ioe) {
      System.out.println(toString() + ".update " + ioe);
    }
  }
  /**
   * Stat update method - used by diskless client.
   * In case of exceptions, sends messages on console.
   * @param jarName archive whose stats will be recorded.
   * @param prop stats to record.
   */
  void statUpdate(String jarName, Properties prop) {
    try {
      ByteArrayOutputStream baos = new ByteArrayOutputStream();
      prop.store(baos, null);
      byte[] buf = baos.toByteArray();
      String jarId = handler.ID + "/" + jarName;
      IcpHeader ih = new IcpHeader(myAddress, ICP_OP_STAT_UPDATE, buf,
        jarId.getBytes("ISO-8859-1"));
      // ICPhandler.dump(ih.getBytes(), ih.length());
      if (handler.bMulticast) {
        DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(), group,
          handler.port);
        mcs.send(hi);
      } else {
        for (int i = 0; i < remAddrs.length; ++i) {
          DatagramPacket hi = new DatagramPacket(ih.getBytes(), ih.length(),
            remAddrs[i], handler.remPorts[i]);
          ds.send(hi);
        }
      }
    }
    catch(UnsupportedEncodingException uee) {
      System.out.println(toString() + ".statUpdate " + uee);
    }
    catch(IOException ioe) {
      System.out.println(toString() + ".statUpdate " + ioe);
    }
  }
  /**
   * Termination:<OL>
   * <LI>Terminates the thread</LI>
   * <LI>Close log files</LI>
   * <LI>Clear logs</LI></OL>
   */
  void close() {
    endLoop();
    if (logs == null)
      return;
    Collection c = logs.values();
    Iterator i = c.iterator();
    while(i.hasNext()) {
      Log l = (Log)i.next();
      l.close();
    }
    logs.clear();
  }
  /** to avoid loitering */
  protected void finalize() {
    close();
  }
}
